// External imports
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;

import java.io.File;
import java.io.IOException;
import java.io.FileInputStream;
import java.io.BufferedInputStream;
import java.util.Random;

import javax.imageio.ImageIO;
import javax.vecmath.Matrix4f;
import javax.vecmath.Point3f;
import javax.vecmath.Vector3f;

import net.java.games.jogl.GLCapabilities;

// Local imports
import org.j3d.aviatrix3d.*;

import org.j3d.aviatrix3d.pipeline.NullCullStage;
import org.j3d.aviatrix3d.pipeline.NullSortStage;
import org.j3d.aviatrix3d.pipeline.StateAndTransparencyDepthSortStage;
import org.j3d.aviatrix3d.pipeline.DefaultRenderPipeline;
import org.j3d.aviatrix3d.pipeline.SingleThreadRenderManager;
import org.j3d.aviatrix3d.surface.DebugAWTSurface;

import org.j3d.util.MatrixUtils;
import org.j3d.geom.CylinderGenerator;
import org.j3d.geom.GeometryData;
import org.j3d.geom.terrain.ElevationGridGenerator;
import org.j3d.texture.procedural.TextureGenerator;

/**
 * Demonstration of water effects using Aviatrix3D and GLSLang shaders.
 * This is a port of the D3D/HLSL demo from the book GPU Gems.
 * <p>
 *
 * @author Justin Couch
 * @version $Revision: 1.1 $
 */
public class WaterRippleDemo extends Frame
    implements ApplicationUpdateObserver,
               NodeUpdateListener,
//               ActionListener,
               WindowListener,
               Runnable
{
    /** pi * 2. Used everywhere here */
    private static final float TWO_PI = (float)Math.PI * 2;

    /** A constant for some form of gravity */
    private static final float GRAV_CONSTANT = 30.f;

    /** Pixel size of the bump map texture */
    private static final int BUMP_TEX_SIZE = 256;

    /** Number of geometry waves to generate */
    private static final int NUM_GEOM_WAVES = 4;

    /** Number of waves to generate on the texture */
    private static final int NUM_TEX_WAVES = 16;

    /** How many bumps should we render in each pass of the wave texture? */
    private static final int BUMPS_PER_PASS = 4;

    /** How many passes should we render? */
    private static final int NUM_BUMP_PASSES  = NUM_TEX_WAVES / BUMPS_PER_PASS;

    /** Size of the basic terrain patch we're dealing with */
    private static final int TERRAIN_SIZE = 100;

    /** Size of the texture, in pixels to create */
    private static final int TEXTURE_SIZE = 512;

    /** Number of points to divide TEXTURE_SIZE by to get grid points */
    private static final int GRID_SPACING = 4;

    /** The texture type constant needed when updating texture coordinates */
    private static final int[] TEX_TYPES =
        { VertexGeometry.TEXTURE_COORDINATE_2 };

    /** Vertex shader file for generating the cosine bump map wave texture */
    private static final String COSINE_VERT_SHADER =
        "shaders/water_cosine.vert";

    /** Fragment shader file for generating the cosine bump map wave texture */
    private static final String COSINE_FRAG_SHADER =
        "shaders/water_cosine.frag";

    /** Vertex shader file for generating the wave surface */
    private static final String WATER_VERT_SHADER =
        "shaders/water_ripple.vert";

    /** Fragment shader file for generating the wave surface */
    private static final String WATER_FRAG_SHADER =
        "shaders/water_ripple.frag";

    /** Names of the files used to create a static environment map */
    private static final String[] WATER_ENV_TEXTURES =
    {
        "textures/water_left_cube_map.jpg",
        "textures/water_right_cube_map.jpg",
        "textures/water_top_cube_map.jpg",
        "textures/water_bottom_cube_map.jpg",
        "textures/water_back_cube_map.jpg",
        "textures/water_front_cube_map.jpg"
    };

    /** Manager for the scene graph handling */
    private SingleThreadRenderManager sceneManager;

    /** Our drawing surface */
    private DrawableSurface surface;

    /** Generator for synthetic textures */
    private TextureGenerator imageGen;

    /** Generator for the terrain */
    private ElevationGridGenerator heightGen;

    /** Data for the generated terrain */
    private GeometryData geomData;

    // Water demo specifics follow
    /** Arguments provided to the water shader */
    private ShaderArguments waterEffectProgram;

    /** Values provided to the cosine shader */
    private ShaderArguments cosineEffectProgram;

    private TextureWaveDescription[] textureWaves;
    private TextureState textureState;
    private GeomWaveDescription[] geomWaves;
    private GeomState geomState;

    /** Common random number generator */
    private Random rgen;

    /** Current view matrix */
    private Matrix4f viewMatrix;

    /** Inverse of the view matrix */
    private Matrix4f positionMatrix;

    /** Time elapse between frames in seconds. */
    private float elapsedTime;

    /** Current time */
    private long lastFrameTime;

    /** The scene graph node to update */
    private TransformGroup rootTransform;

    /** The current angle */
    private float angle;

    /**
     * Construct a new demo with no geometry currently showing, but the
     * default type is set to quads.
     */
    public WaterRippleDemo()
    {
        super("Aviatrix Procedural Water Demo");

        setBackground(SystemColor.menu);
        setLayout(new BorderLayout());
        addWindowListener(this);

//        createParamsPanel();

        textureState = new TextureState();
        geomState = new GeomState();

        textureWaves = new TextureWaveDescription[NUM_TEX_WAVES];

        for(int i = 0; i < NUM_TEX_WAVES; i++)
            textureWaves[i] = new TextureWaveDescription();

        geomWaves = new GeomWaveDescription[NUM_GEOM_WAVES];
        for(int i = 0; i < NUM_GEOM_WAVES; i++)
            geomWaves[i] = new GeomWaveDescription();

        imageGen = new TextureGenerator();
        heightGen = new ElevationGridGenerator(TERRAIN_SIZE,
                                               TERRAIN_SIZE,
                                               TEXTURE_SIZE / GRID_SPACING + 1,
                                               TEXTURE_SIZE / GRID_SPACING + 1);

        geomData = new GeometryData();
        geomData.geometryType = GeometryData.TRIANGLE_STRIPS;
        geomData.geometryComponents = GeometryData.NORMAL_DATA |
                                      GeometryData.TEXTURE_2D_DATA;

        rgen = new Random();
        setupAviatrix();
        setupSceneGraph();

        setSize(600, 600);
        setLocation(40, 40);

        // Need to set visible first before starting the rendering thread due
        // to a bug in JOGL. See JOGL Issue #54 for more information on this.
        // http://jogl.dev.java.net
        setVisible(true);

        Runtime system_runtime = Runtime.getRuntime();
        system_runtime.addShutdownHook(new Thread(this));

        lastFrameTime = System.currentTimeMillis();
    }

    //---------------------------------------------------------------
    // Methods defined by ApplicationUpdateObserver
    //---------------------------------------------------------------

ShaderProgram b_prog;
ShaderProgram w_prog;

ShaderObject b_vert;
ShaderObject b_frag;

ShaderObject w_vert;
ShaderObject w_frag;

    /**
     * Notification that now is a good time to update the scene graph. Use it
     * to clock the particle system.
     */
    public void updateSceneGraph()
    {
        long now = System.currentTimeMillis();

        elapsedTime = (now - lastFrameTime + 1) / 1000.0f;
        lastFrameTime = now;

        updateTexWaves(elapsedTime);
        updateGeomWaves(elapsedTime);

        cosineEffectProgram.dataChanged(this);
        waterEffectProgram.dataChanged(this);
//        rootTransform.boundsChanged(this);

/*
// Print out last logs
System.out.println("bump prog: " + b_prog.getLastInfoLog());
System.out.println("water prog: " + w_prog.getLastInfoLog());

System.out.println("bump vert: " + b_vert.getLastInfoLog());
System.out.println("water vert: " + w_vert.getLastInfoLog());

System.out.println("bump frag: " + b_frag.getLastInfoLog());
System.out.println("water frag: " + w_frag.getLastInfoLog());

// Set these up to log again
b_prog.requestInfoLog();
w_prog.requestInfoLog();

b_vert.requestInfoLog();
w_vert.requestInfoLog();

b_frag.requestInfoLog();
w_frag.requestInfoLog();
*/
    }

    //----------------------------------------------------------
    // Methods required by the UpdateListener interface.
    //----------------------------------------------------------

    /**
     * Notification that its safe to update the node now with any operations
     * that could potentially effect the node's bounds.
     *
     * @param src The node or Node Component that is to be updated.
     */
    public void updateNodeBoundsChanges(Object src)
    {
        angle += Math.PI / 300;

        viewMatrix.setIdentity();
        viewMatrix.rotY(angle);

        rootTransform.setTransform(viewMatrix);
    }

    /**
     * Notification that its safe to update the node now with any operations
     * that only change the node's properties, but do not change the bounds.
     *
     * @param src The node or Node Component that is to be updated.
     */
    public void updateNodeDataChanges(Object src)
    {
        if(src == cosineEffectProgram)
        {
            setCompCosineEffParams(elapsedTime);
        }
        else if(src == waterEffectProgram)
        {
            setWaterParams();
        }
    }

    //---------------------------------------------------------------
    // Methods defined by Runnable
    //---------------------------------------------------------------

    /**
     * Run method for the shutdown hook. This is to deal with someone using
     * ctrl-C to kill the application. Makes sure that all the resources
     * are cleaned up properly.
     */
    public void run()
    {
        shutdownApp();
    }

    //---------------------------------------------------------------
    // Methods defined by WindowListener
    //---------------------------------------------------------------

    /**
     * Ignored
     */
    public void windowActivated(WindowEvent evt)
    {
    }

    /**
     * Ignored
     */
    public void windowClosed(WindowEvent evt)
    {
    }

    /**
     * Exit the application
     *
     * @param evt The event that caused this method to be called.
     */
    public void windowClosing(WindowEvent evt)
    {
        shutdownApp();
        System.exit(0);
    }

    /**
     * Ignored
     */
    public void windowDeactivated(WindowEvent evt)
    {
    }

    /**
     * Ignored
     */
    public void windowDeiconified(WindowEvent evt)
    {
    }

    /**
     * Ignored
     */
    public void windowIconified(WindowEvent evt)
    {
    }

    /**
     * When the window is opened, start everything up.
     */
    public void windowOpened(WindowEvent evt)
    {
        sceneManager.setEnabled(true);
    }

    //---------------------------------------------------------------
    // Local methods
    //---------------------------------------------------------------

    /**
     * Initialise all of the texture waves.
     */
    private void initTexWaves()
    {
        for(int i = 0; i < NUM_TEX_WAVES; i++)
            initTexWave(i);
    }

    /**
     * Initialise a specific wave texture stat set out of the collection of
     * all of them
     *
     * @param waveNum The index of the wave to update
     */
    private void initTexWave(int waveNum)
    {
        float rads = randMinusOneToOne() * textureState.angleDeviation *
                    (float)Math.PI / 180;
        float dx = (float)Math.sin(rads);
        float dy = (float)Math.cos(rads);


        float tx = dx;
        dx = textureState.windDirection[1] * dx - textureState.windDirection[0] * dy;
        dy = textureState.windDirection[0] * tx + textureState.windDirection[1] * dy;

        float maxLen = textureState.maxLength * BUMP_TEX_SIZE /
                       textureState.rippleScale;
        float minLen = textureState.minLength * BUMP_TEX_SIZE /
                       textureState.rippleScale;

        float len = waveNum / (float)(NUM_TEX_WAVES - 1) * (maxLen - minLen) + minLen;

        float reps = BUMP_TEX_SIZE / len;

        dx *= reps;
        dy *= reps;
        dx = (float)((int)(dx >= 0 ? dx + 0.5f : dx - 0.5f));
        dy = (float)((int)(dy >= 0 ? dy + 0.5f : dy - 0.5f));

        textureWaves[waveNum].rotScale[0] = dx;
        textureWaves[waveNum].rotScale[1] = dy;

        float effK = 1 / (float)Math.sqrt(dx * dx + dy * dy);
        len = BUMP_TEX_SIZE * effK;
        textureWaves[waveNum].length = len;
        textureWaves[waveNum].frequency = (float)TWO_PI / len;
        textureWaves[waveNum].amplitude = len * textureState.ampOverLength;
        textureWaves[waveNum].phase = rgen.nextFloat();

        textureWaves[waveNum].direction[0] = dx * effK;
        textureWaves[waveNum].direction[1] = dy * effK;

        textureWaves[waveNum].fade = 1.f;

        float speed = (float)(1 / Math.sqrt(len / (TWO_PI * GRAV_CONSTANT))) / 3;
        speed *= 1 + randMinusOneToOne() * textureState.speedDeviation;
        textureWaves[waveNum].speed = speed;
    }


    /**
     * Update all of the texture waves right now.
     *
     * @param dt The change in time since the last call
     */
    private void updateTexWaves(float dt)
    {
        for(int i = 0; i < NUM_TEX_WAVES; i++ )
            updateTexWave(i, dt);
    }

    /**
     * Update a specific texture wave now.
     *
     * @param waveNum The index of the wave to update
     * @param dt The change in time since the last call
     */
    private void updateTexWave(int waveNum, float dt)
    {
        if(waveNum == textureState.transIdx)
        {
            textureWaves[waveNum].fade += textureState.transDel * dt;

            if(textureWaves[waveNum].fade < 0)
            {
                // This wave is faded out. Re-init and fade it back up.
                initTexWave(waveNum);

                textureWaves[waveNum].fade = 0;
                textureState.transDel = -textureState.transDel;
            }
            else if(textureWaves[waveNum].fade > 1)
            {
                // This wave is faded back up. Start fading another down.
                textureWaves[waveNum].fade = 1;
                textureState.transDel = -textureState.transDel;

                if(++textureState.transIdx >= NUM_TEX_WAVES)
                    textureState.transIdx = 0;
            }
        }

        textureWaves[waveNum].phase -= dt * textureWaves[waveNum].speed;
        textureWaves[waveNum].phase -= (int)textureWaves[waveNum].phase;
    }

    /**
     * Push all the changes through to the bump map shader program now.
     */
    private void setCompCosineEffParams(float dt)
    {
        float[] tmp = new float[4];

        for(int i = 0; i < NUM_TEX_WAVES; i++)
        {
            tmp[0] = textureWaves[i].rotScale[0];
            tmp[1] = textureWaves[i].rotScale[1];
            tmp[2] = 0;
            tmp[3] = textureWaves[i].phase;

            cosineEffectProgram.setUniform("uTrans" + i, 4, tmp, 1);

            float normScale = textureWaves[i].fade / NUM_BUMP_PASSES;
            tmp[0] = textureWaves[i].direction[0] * normScale;
            tmp[1] = textureWaves[i].direction[1] * normScale;
            tmp[2] = 1;
            tmp[3] = 1;

            cosineEffectProgram.setUniform("cCoef" + i, 4, tmp, 1);
        }

        float kRate = 0.1f;
        cosineEffectProgram.getUniform("noiseXform0", tmp);
        tmp[3] += dt * kRate;
        cosineEffectProgram.setUniform("noiseXform0", 4, tmp, 1);

        cosineEffectProgram.getUniform("noiseXform3", tmp);
        tmp[3] += dt * kRate;
        cosineEffectProgram.setUniform("noiseXform3", 4, tmp, 1);

        float s = 0.5f / (BUMPS_PER_PASS + textureState.noise);
        tmp[0] = s;
        tmp[1] = s;
        tmp[2] = 1;
        tmp[3] = 1;

        cosineEffectProgram.setUniform("reScale", 4, tmp, 1);

        float scaleBias = 0.5f * textureState.noise / (NUM_BUMP_PASSES + textureState.noise);
        tmp[0] = scaleBias;
        tmp[1] = scaleBias;
        tmp[2] = 0;
        tmp[3] = 1;

        cosineEffectProgram.setUniform("scaleBias", 4, tmp, 1);
    }

    /**
     * Initialise all the geometry-based waves.
     */
    private void initGeomWaves()
    {
        for(int i = 0; i < NUM_GEOM_WAVES; i++)
            initGeomWave(i);
    }

    private void initGeomWave(int waveNum)
    {
        geomWaves[waveNum].phase = rgen.nextFloat() * TWO_PI;
        geomWaves[waveNum].length = geomState.minLength + rgen.nextFloat() *
                                  (geomState.maxLength - geomState.minLength);
        geomWaves[waveNum].amplitude = geomWaves[waveNum].length *
                                      geomState.ampOverLength / NUM_GEOM_WAVES;
        geomWaves[waveNum].frequency = TWO_PI / geomWaves[waveNum].length;
        geomWaves[waveNum].fade = 1;

        float rotBase = geomState.angleDeviation * (float)Math.PI / 180;

        float rads = rotBase * randMinusOneToOne();
        float rx = (float)Math.cos(rads);
        float ry = (float)Math.sin(rads);

        float x = geomState.windDirection[0];
        float y = geomState.windDirection[1];
        geomWaves[waveNum].direction[0] = x * rx + y * ry;
        geomWaves[waveNum].direction[1] = x * -ry + y * rx;
    }

    /**
     * Update all of the geometry waves right now.
     *
     */
    private void updateGeomWaves(float dt)
    {
        for(int i = 0; i < NUM_GEOM_WAVES; i++)
            updateGeomWave(i, dt);
    }

    private void updateGeomWave(int waveNum, float dt)
    {
        if(waveNum == geomState.transIdx)
        {
            geomWaves[waveNum].fade += geomState.transDel * dt;
            if(geomWaves[waveNum].fade < 0)
            {
                // This wave is faded out. Re-init and fade it back up.
                initGeomWave(waveNum);
                geomWaves[waveNum].fade = 0;
                geomState.transDel = -geomState.transDel;
            }
            else if(geomWaves[waveNum].fade > 1)
            {
                // This wave is faded back up. Start fading another down.
                geomWaves[waveNum].fade = 1;
                geomState.transDel = -geomState.transDel;
                if(++geomState.transIdx >= NUM_GEOM_WAVES)
                    geomState.transIdx = 0;
            }
        }

        float speed = (float)(1 / Math.sqrt(geomWaves[waveNum].length /
                                            (TWO_PI * GRAV_CONSTANT)));

        geomWaves[waveNum].phase += speed * dt;
        geomWaves[waveNum].phase = geomWaves[waveNum].phase % TWO_PI;

        geomWaves[waveNum].amplitude = geomWaves[waveNum].length *
                                       geomState.ampOverLength / NUM_GEOM_WAVES *
                                       geomWaves[waveNum].fade;

    }

    /**
     * Push all the changes through to the bump map shader program now.
     */
    private void setWaterParams()
    {
        float[] tmp = { 0.05f, 0.1f, 0.1f, 0.5f };
        waterEffectProgram.setUniform("waterTint", 4, tmp, 1);

//System.out.println("water tint " + tmp[0] + " " + tmp[1] + " " +tmp[2] + " " + tmp[3]);
        tmp[0] = geomWaves[0].frequency;
        tmp[1] = geomWaves[1].frequency;
        tmp[2] = geomWaves[2].frequency;
        tmp[3] = geomWaves[3].frequency;
        waterEffectProgram.setUniform("cFrequency", 4, tmp, 1);

//System.out.println("freq " + tmp[0] + " " + tmp[1] + " " +tmp[2] + " " + tmp[3]);

        tmp[0] = geomWaves[0].phase;
        tmp[1] = geomWaves[1].phase;
        tmp[2] = geomWaves[2].phase;
        tmp[3] = geomWaves[3].phase;
        waterEffectProgram.setUniform("cPhase", 4, tmp, 1);

//System.out.println("phase " + tmp[0] + " " + tmp[1] + " " +tmp[2] + " " + tmp[3]);

        tmp[0] = geomWaves[0].amplitude;
        tmp[1] = geomWaves[1].amplitude;
        tmp[2] = geomWaves[2].amplitude;
        tmp[3] = geomWaves[3].amplitude;
        waterEffectProgram.setUniform("cAmplitude", 4, tmp, 1);

//System.out.println("amp " + tmp[0] + " " + tmp[1] + " " +tmp[2] + " " + tmp[3]);

        tmp[0] = geomWaves[0].direction[0];
        tmp[1] = geomWaves[1].direction[0];
        tmp[2] = geomWaves[2].direction[0];
        tmp[3] = geomWaves[3].direction[0];
        waterEffectProgram.setUniform("cDirX", 4, tmp, 1);

//System.out.println("cDirX " + tmp[0] + " " + tmp[1] + " " +tmp[2] + " " + tmp[3]);

        tmp[0] = geomWaves[0].direction[1];
        tmp[1] = geomWaves[1].direction[1];
        tmp[2] = geomWaves[2].direction[1];
        tmp[3] = geomWaves[3].direction[1];
        waterEffectProgram.setUniform("cDirY", 4, tmp, 1);

//System.out.println("cDirY " + tmp[0] + " " + tmp[1] + " " +tmp[2] + " " + tmp[3]);

        float normScale = geomState.specAtten * textureState.ampOverLength * TWO_PI;

        normScale *= NUM_BUMP_PASSES + textureState.noise;
        normScale *= textureState.chop + 1;

        tmp[0] = geomState.specEnd;
        tmp[1] = 1 / geomState.specTrans;
        tmp[2] = normScale;
        tmp[3] = 1 / textureState.rippleScale;

        waterEffectProgram.setUniform("cSpecAtten", 4, tmp, 1);

        tmp[0] = positionMatrix.m03;
        tmp[1] = positionMatrix.m13;
        tmp[2] = positionMatrix.m23;
        tmp[3] = 1;
        waterEffectProgram.setUniform("cameraPos", 4, tmp, 1);

//System.out.println("camera pos " + tmp[0] + " " + tmp[1] + " " +tmp[2]);
        tmp[0] = -tmp[0];
        tmp[1] = geomState.envHeight - tmp[1];
        tmp[2] = -tmp[2];

        float g = tmp[0] * tmp[0] + tmp[1] * tmp[1] + tmp[2] * tmp[2] -
                  geomState.envRadius * geomState.envRadius;

        tmp[3] = g;
        waterEffectProgram.setUniform("cEnvAdjust", 4, tmp, 1);

        float K = 5;
        if(geomState.ampOverLength > geomState.chop / (TWO_PI * NUM_GEOM_WAVES * K))
            K = geomState.chop / (TWO_PI * NUM_GEOM_WAVES  * geomState.ampOverLength);

        tmp[0] = geomWaves[0].direction[0] * K;
        tmp[1] = geomWaves[1].direction[0] * K;
        tmp[2] = geomWaves[2].direction[0] * K;
        tmp[3] = geomWaves[3].direction[0] * K;
        waterEffectProgram.setUniform("cDirXK", 4, tmp, 1);

//System.out.println("cDirXK " + tmp[0] + " " + tmp[1] + " " +tmp[2] + " " + tmp[3]);
        tmp[0] = geomWaves[0].direction[1] * K;
        tmp[1] = geomWaves[1].direction[1] * K;
        tmp[2] = geomWaves[2].direction[1] * K;
        tmp[3] = geomWaves[3].direction[1] * K;

        waterEffectProgram.setUniform("cDirYK", 4, tmp, 1);
//System.out.println("cDirYK " + tmp[0] + " " + tmp[1] + " " +tmp[2] + " " + tmp[3]);

        tmp[0] = geomWaves[0].direction[0] * geomWaves[0].frequency;
        tmp[1] = geomWaves[1].direction[0] * geomWaves[1].frequency;
        tmp[2] = geomWaves[2].direction[0] * geomWaves[2].frequency;
        tmp[3] = geomWaves[3].direction[0] * geomWaves[3].frequency;
        waterEffectProgram.setUniform("cDirXW", 4, tmp, 1);

//System.out.println("cDirXW " + tmp[0] + " " + tmp[1] + " " +tmp[2] + " " + tmp[3]);

        tmp[0] = geomWaves[0].direction[1] * geomWaves[0].frequency;
        tmp[1] = geomWaves[1].direction[1] * geomWaves[1].frequency;
        tmp[2] = geomWaves[2].direction[1] * geomWaves[2].frequency;
        tmp[3] = geomWaves[3].direction[1] * geomWaves[3].frequency;
        waterEffectProgram.setUniform("cDirYW", 4, tmp, 1);

//System.out.println("cDirYW " + tmp[0] + " " + tmp[1] + " " +tmp[2] + " " + tmp[3]);

        tmp[0] = K * geomWaves[0].frequency;
        tmp[1] = K * geomWaves[1].frequency;
        tmp[2] = K * geomWaves[2].frequency;
        tmp[3] = K * geomWaves[3].frequency;
        waterEffectProgram.setUniform("cKW", 4, tmp, 1);

//System.out.println("cKW " + tmp[0] + " " + tmp[1] + " " +tmp[2] + " " + tmp[3]);

        tmp[0] = geomWaves[0].direction[0] * geomWaves[0].direction[0] * K * geomWaves[0].frequency;
        tmp[1] = geomWaves[1].direction[0] * geomWaves[1].direction[0] * K * geomWaves[1].frequency;
        tmp[2] = geomWaves[2].direction[0] * geomWaves[2].direction[0] * K * geomWaves[2].frequency;
        tmp[3] = geomWaves[3].direction[0] * geomWaves[3].direction[0] * K * geomWaves[3].frequency;
        waterEffectProgram.setUniform("cDirXSqKW", 4, tmp, 1);

        tmp[0] = geomWaves[0].direction[1] * geomWaves[0].direction[1] * K * geomWaves[0].frequency;
        tmp[1] = geomWaves[1].direction[1] * geomWaves[1].direction[1] * K * geomWaves[1].frequency;
        tmp[2] = geomWaves[2].direction[1] * geomWaves[2].direction[1] * K * geomWaves[2].frequency;
        tmp[3] = geomWaves[3].direction[1] * geomWaves[3].direction[1] * K * geomWaves[3].frequency;
        waterEffectProgram.setUniform("cDirYSqKW", 4, tmp, 1);

        tmp[0] = geomWaves[0].direction[0] * geomWaves[0].direction[1] * K * geomWaves[0].frequency;
        tmp[1] = geomWaves[1].direction[0] * geomWaves[1].direction[1] * K * geomWaves[1].frequency;
        tmp[2] = geomWaves[2].direction[0] * geomWaves[2].direction[1] * K * geomWaves[2].frequency;
        tmp[3] = geomWaves[3].direction[0] * geomWaves[3].direction[1] * K * geomWaves[3].frequency;
        waterEffectProgram.setUniform("cDirXDirYKW", 4, tmp, 1);
    }

    /**
     * Setup the avaiatrix pipeline here.
     */
    private void setupAviatrix()
    {
        // Assemble a simple single-threaded pipeline.
        GLCapabilities caps = new GLCapabilities();
        caps.setDoubleBuffered(true);
        caps.setHardwareAccelerated(true);

        CullStage culler = new NullCullStage();
        culler.setOffscreenCheckEnabled(false);

        SortStage sorter = new StateAndTransparencyDepthSortStage();
        surface = new DebugAWTSurface(caps);
        surface.setClearColor(0, 0, 0.2f, 1);
        DefaultRenderPipeline pipeline = new DefaultRenderPipeline();

        pipeline.setCuller(culler);
        pipeline.setSorter(sorter);
        pipeline.setDrawableSurface(surface);

        // Render manager
        sceneManager = new SingleThreadRenderManager();
        sceneManager.addPipeline(pipeline);
        sceneManager.setDrawableSurface(surface);
        sceneManager.setMinimumFrameInterval(30);

        // Before putting the pipeline into run mode, put the canvas on
        // screen first.
        Component comp = (Component)surface.getSurfaceObject();
        add(comp, BorderLayout.CENTER);
    }

    /**
     * Setup the basic scene which consists of a quad and a viewpoint
     *
     * @param nSphere Number of sphere objects to pre-initialise
     */
    private void setupSceneGraph()
    {
        viewMatrix = new Matrix4f();
        positionMatrix = new Matrix4f();

        MatrixUtils mat_utilts = new MatrixUtils();
        Point3f eye = new Point3f(0, 20, 100);
        Point3f center = new Point3f(0, 0, 0);
        Vector3f up = new Vector3f(0, 1, 0);
        MatrixUtils mat_utils = new MatrixUtils();
        mat_utils.lookAt(eye, center, up, viewMatrix);
        mat_utils.inverse(viewMatrix, positionMatrix);

        // View group
        float[] global_color = { 240.0f / 255, 240.0f / 255, 240.0f / 255};
        Viewpoint vp = new Viewpoint();
        vp.setHeadlightEnabled(true);
        vp.setGlobalAmbientLightEnabled(true);
        vp.setGlobalAmbientColor(global_color);

        // We could set these directly, but we don't because we'd like to
        // pass the values through to the shaders as well. More convenient and
        // we guarantee the same values then.
        TransformGroup tx = new TransformGroup();
        tx.setTransform(positionMatrix);
        tx.addChild(vp);

        TransformGroup main_scene_root = new TransformGroup();
//        Matrix4f mat = new Matrix4f();
//        mat.setIdentity();
//        main_scene_root.setTransform(mat);
        main_scene_root.addChild(tx);

        rootTransform = new TransformGroup();
        main_scene_root.addChild(rootTransform);

        DirectionalLight world_light = new DirectionalLight();
        world_light.setDirection(-1, -2, -1);
        world_light.setEnabled(true);
        rootTransform.addChild(world_light);

        Group poles = createPoles();
        rootTransform.addChild(poles);

        Group terrain = createTerrain();
        rootTransform.addChild(terrain);

        Group water = createWater();
        rootTransform.addChild(water);

        Scene main_scene = new Scene();
        main_scene.setRenderedGeometry(main_scene_root);
        main_scene.setActiveView(vp);

        sceneManager.setApplicationObserver(this);
        sceneManager.setScene(main_scene);
    }

    /**
     * Create the new water mesh. In the original demo this loaded it directly
     * from a file of vertices and then generates a mesh from it. For this demo,
     * we're just going to assume a single flat, rectangular area and tesselate
     * it appropriately into a flat mesh.
     *
     * Also, spend some time loading the needed shader.
     */
    private Group createWater()
    {
        // Start with the pBuffer texture. This is used as the bumpmap source.
        // To get a bumpmap that covers the whole texture, we need to draw a
        // really big quad to it and use the shader on that quad to generate
        // the basic map. Set the camera to orthographic mode and that will take
        // care of filling up the entire screen.

        ShaderObject vert_shader = new ShaderObject(true);
        vert_shader.setSourceStrings(loadShaderSource(COSINE_VERT_SHADER), 1);
        vert_shader.compile();

        ShaderObject frag_shader = new ShaderObject(false);
        frag_shader.setSourceStrings(loadShaderSource(COSINE_FRAG_SHADER), 1);
        frag_shader.compile();

        ShaderProgram prog = new ShaderProgram();
        prog.addShaderObject(vert_shader);
        prog.addShaderObject(frag_shader);
        prog.link();

b_prog = prog;
b_vert = vert_shader;
b_frag = frag_shader;

        cosineEffectProgram = new ShaderArguments();
        setupBumpMapConstants();

        GLSLangShader bump_shader = new GLSLangShader();
        bump_shader.setShaderProgram(prog);
        bump_shader.setShaderArguments(cosineEffectProgram);

        // Texture units to hold the bias noise and cosineLUT textures
        TextureUnit[] bump_texs = new TextureUnit[2];
        bump_texs[0] = new TextureUnit();
        bump_texs[1] = new TextureUnit();
        bump_texs[0].setTexture(createCosineLUT());
        bump_texs[1].setTexture(createBiasNoiseMap());

        BlendAttributes ba = new BlendAttributes();
        ba.setDestinationBlendFactor(BlendAttributes.BLEND_ZERO);

// Not sure what this corresponds to here
//            m_pd3dDevice->SetRenderState(D3DRS_ALPHAFUNC, D3DCMP_ALWAYS);

        Appearance app = new Appearance();
        app.setShader(bump_shader);
        app.setTextureUnits(bump_texs, 2);
        app.setBlendAttributes(ba);

        VertexGeometry geom = createClearBuffer();

        Shape3D b_shape = new Shape3D();
        b_shape.setGeometry(geom);
        b_shape.setAppearance(app);

        Viewpoint bump_vp = new Viewpoint();

        Group water_root = new Group();
        water_root.addChild(b_shape);
        water_root.addChild(bump_vp);

        ViewEnvironment view_env = new ViewEnvironment();
        view_env.setProjectionType(ViewEnvironment.ORTHOGRAPHIC_PROJECTION);

        Scene bump_scene = new Scene();
        bump_scene.setRenderedGeometry(water_root);
        bump_scene.setViewEnvironment(view_env);
        bump_scene.setActiveView(bump_vp);

        GLCapabilities caps = new GLCapabilities();
        caps.setDoubleBuffered(false);
        caps.setHardwareAccelerated(true);

        OffscreenTexture2D bump_texture =
            new OffscreenTexture2D(caps, BUMP_TEX_SIZE, BUMP_TEX_SIZE);
        bump_texture.setRepaintRequired(true);
        bump_texture.setRenderedScene(bump_scene);

        // Ok, now, on with the rest of the show. We need to create a water
        // grid of points to render, and shaders to deal with it. One of the
        // inputs to the shader is the bump map pBuffer we've just created.
        // Since this is a demo, we just create a rectangular grid of tristrips
        // though in a real app we could create any arbitrary geometry
        // structure here, not even a grid.
        vert_shader = new ShaderObject(true);
        vert_shader.setSourceStrings(loadShaderSource(WATER_VERT_SHADER), 1);
        vert_shader.compile();

        frag_shader = new ShaderObject(false);
        frag_shader.setSourceStrings(loadShaderSource(WATER_FRAG_SHADER), 1);
        frag_shader.compile();

        prog = new ShaderProgram();
        prog.addShaderObject(vert_shader);
        prog.addShaderObject(frag_shader);
        prog.link();

w_prog = prog;
w_vert = vert_shader;
w_frag = frag_shader;

        waterEffectProgram = new ShaderArguments();
        resetWater();
        setupWaterParamConstants();

        GLSLangShader water_shader = new GLSLangShader();
        water_shader.setShaderProgram(prog);
        water_shader.setShaderArguments(waterEffectProgram);

        TextureUnit[] water_texs = new TextureUnit[2];
        water_texs[0] = new TextureUnit();
        water_texs[0].setTexture(bump_texture);
        water_texs[1] = loadEnvMap();

        ba = new BlendAttributes();
        PolygonAttributes pa = new PolygonAttributes();
        pa.setDrawMode(true, PolygonAttributes.DRAW_LINE);

        app = new Appearance();
        app.setShader(water_shader);
        app.setTextureUnits(water_texs, 2);
        app.setBlendAttributes(ba);
//        app.setPolygonAttributes(pa);

        geom = createWaterGrid();

        Shape3D w_shape = new Shape3D();
        w_shape.setGeometry(geom);
        w_shape.setAppearance(app);

        Group main_group = new Group();
        main_group.addChild(w_shape);

        return main_group;
    }

    /**
     * Convenience method to create just a flat grid of water points and
     * return the geometry description thereof.
     */
    private VertexGeometry createWaterGrid()
    {
        float width = TERRAIN_SIZE;
        float depth = TERRAIN_SIZE;
        int nWPnts = 25;
        int nDPnts = 25;

        float[] coords = new float[nWPnts * nDPnts * 3 * 2];
        float[] colors = new float[nWPnts * nDPnts * 3 * 2];
        int[] strips = new int[nDPnts - 1];
        float w_inc = width / (nWPnts - 1);
        float d_inc = depth / (nDPnts - 1);
        float w = -width * 0.5f;
        float d = -depth * 0.5f;

        float dist = (float)Math.sqrt(w_inc * w_inc + d_inc * d_inc);

        int idx = 0;

        for(int y = 0; y < nDPnts - 1; y++)
        {
            for(int x = 0; x < nWPnts; x++)
            {
                 coords[idx] = w;
                 coords[idx + 2] = d;

                 coords[idx + 3] = w;
                 coords[idx + 5] = d + d_inc;

                 colors[idx] = dist;
                 colors[idx + 3] = dist;

                 idx += 6;

                 w += w_inc;
            }

            w = -width * 0.5f;
            d += d_inc;


            strips[y] = nWPnts * 2;
        }

        TriangleStripArray geom = new TriangleStripArray();
        geom.setVertices(VertexGeometry.COORDINATE_3,
                         coords,
                         nWPnts * nDPnts * 2);
        geom.setStripCount(strips, nDPnts - 1);
        geom.setColors(false, colors);

        return geom;
    }

    /**
     * Create the geometry and structure for the poles. Creates a random number
     * of poles between 1 and 10 and places them evenly in a circle around the
     * width of the terrain.
     */
    private Group createPoles()
    {
        // Create the geometry to be shared amongst all of them.
        CylinderGenerator cyl_gen = new CylinderGenerator(20, 1);
        GeometryData data = new GeometryData();
        data.geometryType = GeometryData.TRIANGLE_STRIPS;
        data.geometryComponents = GeometryData.NORMAL_DATA;
        cyl_gen.generate(data);

        TriangleStripArray geom = new TriangleStripArray();
        geom.setVertices(VertexGeometry.COORDINATE_3,
                         data.coordinates,
                         data.vertexCount);
        geom.setStripCount(data.stripCounts, data.numStrips);
        geom.setNormals(data.normals);

        Material mat = new Material();
        mat.setDiffuseColor(new float[] { 0.5f, 0.3f, 0 });
        mat.setEmissiveColor(new float[] { 0.2f, 0.2f, 0.2f });
/*
        ShaderObject frag_shader = new ShaderObject(false);
        frag_shader.setSourceStrings(loadShaderSource(WOOD_FRAG_SHADER));
        frag_shader.compile();

        ShaderProgram prog = new ShaderProgram();
        prog.addShaderObject(frag_shader);
        prog.link();

        ShaderArguments args = new ShaderArguments();

        GLSLangShader shader = new GLSLangShader();
        shader.setShaderProgram(prog);
*/

        Appearance app = new Appearance();
        app.setMaterial(mat);
//        app.setShader(shader);

        Shape3D shape = new Shape3D();
        shape.setGeometry(geom);
        shape.setAppearance(app);

        SharedNode shared_pole = new SharedNode();
        shared_pole.setChild(shape);

        int num_poles = (int)(rgen.nextFloat() * 10);
        float angle_inc = TWO_PI / num_poles;
        float angle = 0;

        Group pole_group = new Group();
        Matrix4f loc_mat = new Matrix4f();
        loc_mat.setIdentity();
        loc_mat.m13 = 5; // shift up a bit, but not all the way

        for(int i = 0; i < num_poles; i++)
        {
            float x = (float)Math.sin(angle) * TERRAIN_SIZE * 0.5f;
            float y = (float)Math.cos(angle) * TERRAIN_SIZE * 0.5f;
            angle += angle_inc;

            loc_mat.m03 = x;
            loc_mat.m23 = y;

            TransformGroup tg = new TransformGroup();
            tg.setTransform(loc_mat);
            tg.addChild(shared_pole);

            pole_group.addChild(tg);
        }

        return pole_group;
    }

    /**
     * Create the terrain section.
     */
    private Group createTerrain()
    {
        // Create a small random set of terrain and a texture to go onto it.
        int width = TEXTURE_SIZE;
        int depth = TEXTURE_SIZE;
        int passes = 5;
        float scale = 5;
        float freq = 4;
        float y_scale = 0.3f;
        float freq_diff = 6;

        float[] raw_img = new float[width * depth];

        imageGen.generateSynthesisTexture(raw_img,
                                          freq,
                                          scale,
                                          width,
                                          depth);

        for(int i = 1; i < passes; i++)
        {
            freq += freq_diff;
            scale *= y_scale;
            imageGen.generateSynthesisTexture(raw_img,
                                              freq,
                                              scale,
                                              width,
                                              depth);
        }

        // find min and max values of the floats
        float min_y = raw_img[0];
        float max_y = raw_img[0];

        for(int i = 0; i < width * depth; i++)
        {
            if(raw_img[i] > max_y)
                max_y = raw_img[i];
            else if(raw_img[i] < min_y)
                min_y = raw_img[i];
        }

        float diff = 1 / (max_y - min_y);

        for(int i = 0; i < width * depth; i++)
        {
//            pixels[i] = (byte)(((raw_img[i] - min_y) * diff) * 255);
//            pixel_heights[i] = ((raw_img[i] - min_y) * diff) * 255;
            raw_img[i] = raw_img[i] * 0.5f - 5;
        }

        // Now generate the geometry updates. First need to down-sample the
        // height map image. A 512x512 source image equates to 500K
        // triangles. Take every 4th height
        int sample_size = (TEXTURE_SIZE / GRID_SPACING + 1);
        float[] terrain_heights = new float[sample_size * sample_size];
        int t_idx = 0;
        for(int y = 0; y < TEXTURE_SIZE / GRID_SPACING; y++)
        {
            for(int x = 0; x < TEXTURE_SIZE / GRID_SPACING; x++)
                terrain_heights[t_idx++] = raw_img[y * width * GRID_SPACING + x * GRID_SPACING];

            // Then pick up the one at the end
            terrain_heights[t_idx++] = raw_img[y * width * GRID_SPACING + width - 1];
        }

        // Then pick up the one at the end
        for(int x = 0; x < TEXTURE_SIZE / GRID_SPACING; x++)
            terrain_heights[t_idx++] = raw_img[width * (depth - 1) + x * GRID_SPACING];

        terrain_heights[t_idx++] = raw_img[TEXTURE_SIZE * TEXTURE_SIZE - 1];

        heightGen.setTerrainDetail(terrain_heights, 0);
        heightGen.generate(geomData);

        TriangleStripArray geom = new TriangleStripArray();
        geom.setVertices(VertexGeometry.COORDINATE_3,
                         geomData.coordinates,
                         geomData.vertexCount);
        geom.setStripCount(geomData.stripCounts,
                                      geomData.numStrips);
        geom.setNormals(geomData.normals);

        float[][] tex_coords = { geomData.textureCoordinates };
        geom.setTextureCoordinates(TEX_TYPES, tex_coords, 1);

        Material mat = new Material();
        mat.setDiffuseColor(new float[] { 0, 1, 1 });
/*
        ShaderObject frag_shader = new ShaderObject(false);
        frag_shader.setSourceStrings(loadShaderSource(SAND_FRAG_SHADER));
        frag_shader.compile();

        ShaderProgram prog = new ShaderProgram();
        prog.addShaderObject(frag_shader);
        prog.link();

        ShaderArguments args = new ShaderArguments();

        GLSLangShader shader = new GLSLangShader();
        shader.setShaderProgram(prog);
*/

        Appearance app = new Appearance();
        app.setMaterial(mat);
//        app.setShader(shader);

        Shape3D shape = new Shape3D();
        shape.setGeometry(geom);
        shape.setAppearance(app);

        Group terrain_group = new Group();
        terrain_group.addChild(shape);

        return terrain_group;
    }

    /**
     * Load a static environment map for use as a cubic environment map
     * for the water refraction effects.
     */
    private TextureUnit loadEnvMap()
    {
        // Load the texture image
        TextureComponent2D[] img_sides = new TextureComponent2D[6];

        for(int i = 0; i < 6; i++)
            img_sides[i] = loadImage(new File(WATER_ENV_TEXTURES[i]));

        TextureCubicEnvironmentMap texture = new TextureCubicEnvironmentMap();
        texture.setImages(Texture.MODE_BASE_LEVEL,
                          Texture.FORMAT_RGB,
                          img_sides,
                          6);

/*
        TexCoordGeneration coord_gen = new TexCoordGeneration();
        coord_gen.setParameter(TexCoordGeneration.TEXTURE_S,
                               TexCoordGeneration.MODE_GENERIC,
//                               TexCoordGeneration.MAP_NORMALS,
                               TexCoordGeneration.MAP_REFLECTIONS,
                               null);

        coord_gen.setParameter(TexCoordGeneration.TEXTURE_T,
                               TexCoordGeneration.MODE_GENERIC,
//                               TexCoordGeneration.MAP_NORMALS,
                               TexCoordGeneration.MAP_REFLECTIONS,
                               null);

        coord_gen.setParameter(TexCoordGeneration.TEXTURE_R,
                               TexCoordGeneration.MODE_GENERIC,
//                               TexCoordGeneration.MAP_NORMALS,
                               TexCoordGeneration.MAP_REFLECTIONS,
                               null);
*/

        TextureUnit tu = new TextureUnit();
        tu.setTexture(texture);
//        tu.setTexCoordGeneration(coord_gen);

        return tu;
    }

    /**
     * Reset the complete water state back to time zero.
     */
    private void resetWater()
    {
        initTextureState();
        initGeomState();

        initTexWaves();
        initGeomWaves();
    }

    /**
     * Initialise the global geometry wave state.
     */
    private void initGeomState()
    {
        geomState.chop = 2.5f;
        geomState.angleDeviation = 15;
        geomState.windDirection[0] = 0;
        geomState.windDirection[1] = 1;

        geomState.minLength = 15;
        geomState.maxLength = 25;
        geomState.ampOverLength = 0.6f;

        geomState.envHeight = -50;
        geomState.envRadius = 100;
        geomState.waterLevel = -2;

        geomState.transIdx = 0;
        geomState.transDel = -1 / 10.0f;

        geomState.specAtten = 1;
        geomState.specEnd = 200;
        geomState.specTrans = 100;
    }

    /**
     * Initialise the basic state common to all textures. These can be later
     * modified by UI inputs.
     */
    private void initTextureState()
    {
        textureState.noise = 0.2f;
        textureState.chop = 1;
        textureState.angleDeviation = 15;
        textureState.windDirection[0] = 0;
        textureState.windDirection[1] = 1.f;
        textureState.maxLength = 10;
        textureState.minLength = 1;
        textureState.ampOverLength = 0.1f;
        textureState.rippleScale = 25;
        textureState.speedDeviation = 0.1f;

        textureState.transIdx = 0;
        textureState.transDel = -0.2f;
    }

    /**
     * The clear buffer is just a quad, with a texture on it so that we can
     * use the shaders to render to the quad and hav it apply over the whole
     * surface.
     */
    private VertexGeometry createClearBuffer()
    {
        float[] coords = new float[12];
        float[][] tex_coords = new float[1][8];

        coords[0] = -1;
        coords[1] = -1;
        coords[2] = -0.5f;

        coords[3] = 1;
        coords[4] = -1;
        coords[5] = -0.5f;

        coords[6] = 1;
        coords[7] = 1;
        coords[8] = -0.5f;

        coords[9] = -1;
        coords[10] = 1;
        coords[11] = -0.5f;

        // Yes, these are offset by 1/2. Not sure why. The original code
        // has it this way too.
        tex_coords[0][0] = 0.5f / BUMP_TEX_SIZE;
        tex_coords[0][1] = 0.5f / BUMP_TEX_SIZE;

        tex_coords[0][2] = 1 + tex_coords[0][0];
        tex_coords[0][3] = tex_coords[0][1];

        tex_coords[0][4] = 1 + tex_coords[0][0];
        tex_coords[0][5] = 1 + tex_coords[0][1];

        tex_coords[0][6] = tex_coords[0][0];
        tex_coords[0][7] = 1 + tex_coords[0][1];

        QuadArray geom = new QuadArray();
        geom.setVertices(QuadArray.COORDINATE_3, coords, 4);
        geom.setTextureCoordinates(TEX_TYPES, tex_coords, 1);

        return geom;
    }

    /**
     * Create the cosine lookup texture. This is a 1D texture with the cosine
     * and sine values encoded as 8-bit values in the R and G components of the
     * texture. Potentially this could be reduced to a 2-component I-A texture
     * rather than the 4-component version currently in use.
     * The created texture is left in the cosineLUT class var.
     */
    private Texture createCosineLUT()
    {
        byte[] pixels = new byte[BUMP_TEX_SIZE * 4];
        float i_size = 1.0f / (BUMP_TEX_SIZE - 1);

        for(int i = 0; i < BUMP_TEX_SIZE; i++ )
        {
            float dist = i * i_size  * TWO_PI;
            float c = (float)Math.cos(dist);
            float s = (float)Math.sin(dist);
            s *= 0.5f;
            s += 0.5f;
            s = (float)Math.pow(s, textureState.chop);
            c *= s;
            byte cos_dist = (byte)((c * 0.5f + 0.5f) * 255.999f);
            pixels[i * 4] = 0;
            pixels[i * 4 + 1] = cos_dist;
            pixels[i * 4 + 2] = cos_dist;
            pixels[i * 4 + 3] = 0;
        }

        ByteTextureComponent1D tex_comp =
            new ByteTextureComponent1D(TextureComponent.FORMAT_RGBA,
                                       BUMP_TEX_SIZE,
                                       pixels);

        return new Texture1D(Texture.FORMAT_RGBA, tex_comp);
    }

    /**
     * Create the noise map for generating bias values in the wave normal
     * map. The bias values are encoded in the R and G components of the
     * texture with A and B left as zeroes. Will see if we can reduce this
     * to RGB or IA-only textures.
     */
    private Texture createBiasNoiseMap()
    {
        byte[] pixels = new byte[BUMP_TEX_SIZE * BUMP_TEX_SIZE * 4];

        for(int i = 0; i < BUMP_TEX_SIZE * BUMP_TEX_SIZE; i++)
        {
            float x = rgen.nextFloat();
            float y = rgen.nextFloat();

            byte r = (byte)(x * 255.999f);
            byte g = (byte)(y * 255.999f);

            pixels[i * 4] = 0;
            pixels[i * 4 + 1] = r;
            pixels[i * 4 + 2] = r;
            pixels[i * 4 + 3] = 0;
        }

        ByteTextureComponent2D tex_comp =
            new ByteTextureComponent2D(TextureComponent.FORMAT_RGBA,
                                       BUMP_TEX_SIZE,
                                       BUMP_TEX_SIZE,
                                       pixels);

        return new Texture2D(Texture.FORMAT_RGBA, tex_comp);
    }

    /**
     * Set up the constant uniform values that don't change over time
     * for the water renderer.
     */
    private void setupWaterParamConstants()
    {
        float[] tmp = new float[4];
        tmp[0] = geomWaves[0].length;
        tmp[1] = geomWaves[1].length;
        tmp[2] = geomWaves[2].length;
        tmp[3] = geomWaves[3].length;
        waterEffectProgram.setUniform("cLengths", 4, tmp, 1);

        tmp[0] = geomState.waterLevel + 1;
        tmp[1] = geomState.waterLevel + 1;
        tmp[2] = geomState.waterLevel;
        tmp[3] = geomState.waterLevel;
        waterEffectProgram.setUniform("depthOffset", 4, tmp, 1);

        tmp[0] = 0.5f;
        tmp[0] = 0.5f;
        tmp[0] = 0.5f;
        tmp[0] = 1;
        waterEffectProgram.setUniform("cDepthScale", 4, tmp, 1);

        tmp[0] = -200;
        tmp[1] = 0.01f;
        tmp[2] = 0;
        tmp[3] = 1;
        waterEffectProgram.setUniform("cFogParams", 4, tmp, 1);

        waterEffectProgram.setUniformSampler("bumpMap", 0);
        waterEffectProgram.setUniformSampler("envMap", 1);
    }

    /**
     * Initialise the uniform variables for the shader used in the
     * bump map rendering.
     */
    private void setupBumpMapConstants()
    {
        float[] init = { 20, 0, 0, 0 };
        cosineEffectProgram.setUniform("noiseXform0", 4, init, 1);
        cosineEffectProgram.setUniform("noiseXform2", 4, init, 1);

        init[0] = 0;
        init[1] = 20;

        cosineEffectProgram.setUniform("noiseXform1", 4, init, 1);
        cosineEffectProgram.setUniform("noiseXform3", 4, init, 1);

        cosineEffectProgram.setUniformSampler("cosineLUT", 0);
        cosineEffectProgram.setUniformSampler("biasNoise", 1);
    }

    /**
     * Load a single image.
     */
    private TextureComponent2D loadImage(File f)
    {
        ImageTextureComponent2D img_comp = null;

        try
        {
            if(!f.exists())
                System.out.println("Can't find texture source file");

            FileInputStream is = new FileInputStream(f);

            BufferedInputStream stream = new BufferedInputStream(is);
            BufferedImage img = ImageIO.read(stream);

            int img_width = img.getWidth();
            int img_height = img.getHeight();
            int format = TextureComponent.FORMAT_RGB;

            switch(img.getType())
            {
                case BufferedImage.TYPE_3BYTE_BGR:
                case BufferedImage.TYPE_CUSTOM:
                case BufferedImage.TYPE_INT_RGB:
                    break;

                case BufferedImage.TYPE_4BYTE_ABGR:
                case BufferedImage.TYPE_INT_ARGB:
                    format = TextureComponent.FORMAT_RGBA;
                    break;
            }

            img_comp = new ImageTextureComponent2D(format,
                                            img_width,
                                            img_height,
                                            img);

        }
        catch(IOException ioe)
        {
            System.out.println("Error reading image: " + ioe);
        }

        return img_comp;
    }

    /**
     * Load a shader source file into an array of strings.
     *
     * @param name The name of the file to load
     * @return the strings that represent the source.
     */
    private String[] loadShaderSource(String file)
    {
        File f = new File(file);
        if(!f.exists())
        {
            System.out.println("Can't find shader " + file);
            return null;
        }

        String str = null;

        try
        {
            FileInputStream fis = new FileInputStream(f);
            byte[] raw_chars = new byte[(int)f.length()];
            byte[] readbuf = new byte[1024];
            int bytes_read = 0;
            int read_offset = 0;

            while((bytes_read = fis.read(readbuf, 0, 1024)) != -1)
            {
                System.arraycopy(readbuf, 0, raw_chars, read_offset, bytes_read);
                read_offset += bytes_read;
            }

                str = new String(raw_chars);
        }
        catch(IOException ioe)
        {
            System.out.println("error reading shader file " + ioe.getMessage());
            return null;
        }

        return new String[] { str };
    }

    /**
     * Generate an inverted random number in the range [-1, 1]
     *
     * @return A new random number
     */
    private float randMinusOneToOne()
    {
        return rgen.nextFloat() * 2.0f - 1;
    }

    /**
     * Close down the application safely by destroying all the resources
     * currently in use.
     */
    private void shutdownApp()
    {
        sceneManager.setEnabled(false);
        surface.dispose();
    }

    public static void main(String[] argv)
    {
        WaterRippleDemo demo = new WaterRippleDemo();
        demo.setVisible(true);
    }
}