
// External imports
import java.awt.*;
import java.awt.event.*;

import java.awt.image.BufferedImage;
import javax.imageio.ImageIO;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.FileInputStream;
import java.io.BufferedInputStream;

import javax.vecmath.Matrix4f;
import javax.vecmath.Vector3f;

import net.java.games.jogl.GLCapabilities;

// Local imports
import org.j3d.aviatrix3d.*;

import org.j3d.aviatrix3d.pipeline.NullCullStage;
import org.j3d.aviatrix3d.pipeline.NullSortStage;
import org.j3d.aviatrix3d.pipeline.DefaultRenderPipeline;
import org.j3d.aviatrix3d.pipeline.SingleThreadRenderManager;
import org.j3d.aviatrix3d.surface.DebugAWTSurface;

import org.j3d.geom.GeometryData;
import org.j3d.geom.SphereGenerator;

/**
 * Example application that demonstrates using the simple GLSLang texture
 * shader from Chapter 10 of the Orange Book.
 *
 * @author Justin Couch
 * @version $Revision: 1.1 $
 */
public class OBToyBallDemo extends Frame
    implements WindowListener, Runnable
{
    /** Vertex shader source file */
    private static final String VTX_SHADER_FILE =
        "orangebook/CH11-toyball.vert";

    /** Fragment shader source file */
    private static final String FRAG_SHADER_FILE =
        "orangebook/CH11-toyball.frag";

    /** Manager for the scene graph handling */
    private SingleThreadRenderManager sceneManager;

    /** Our drawing surface */
    private DrawableSurface surface;

    /** The shader for vertex section */
    private VertexShader vtxShader;

    /** The shader for fragment processing */
    private FragmentShader fragShader;

    /**
     * Construct a new shader demo instance.
     */
    public OBToyBallDemo()
    {
        super("Demo Illustrating OpenGL Shaders from the orange book");

        setLayout(new BorderLayout());
        addWindowListener(this);

        setupAviatrix();
        setupSceneGraph();

        setSize(600, 600);
        setLocation(40, 40);

        // Need to set visible first before starting the rendering thread due
        // to a bug in JOGL. See JOGL Issue #54 for more information on this.
        // http://jogl.dev.java.net
        setVisible(true);

        Runtime system_runtime = Runtime.getRuntime();
        system_runtime.addShutdownHook(new Thread(this));
    }

    /**
     * Setup the avaiatrix pipeline here
     */
    private void setupAviatrix()
    {
        // Assemble a simple single-threaded pipeline.
        GLCapabilities caps = new GLCapabilities();
        caps.setDoubleBuffered(true);
        caps.setHardwareAccelerated(true);

        CullStage culler = new NullCullStage();
        culler.setOffscreenCheckEnabled(false);

        SortStage sorter = new NullSortStage();
        surface = new DebugAWTSurface(caps);
        DefaultRenderPipeline pipeline = new DefaultRenderPipeline();

        pipeline.setCuller(culler);
        pipeline.setSorter(sorter);
        pipeline.setDrawableSurface(surface);

        // Render manager
        sceneManager = new SingleThreadRenderManager();
        sceneManager.addPipeline(pipeline);
        sceneManager.setDrawableSurface(surface);
        sceneManager.setMinimumFrameInterval(100);

        // Before putting the pipeline into run mode, put the canvas on
        // screen first.
        Component comp = (Component)surface.getSurfaceObject();
        add(comp, BorderLayout.CENTER);
    }

    /**
     * Setup the basic scene which consists of a quad and a viewpoint
     */
    private void setupSceneGraph()
    {
        // View group
        Viewpoint vp = new Viewpoint();

        Vector3f trans = new Vector3f(0, 0, 3f);

        Matrix4f mat = new Matrix4f();
        mat.setIdentity();
        mat.setTranslation(trans);

        TransformGroup tx = new TransformGroup();
        tx.addChild(vp);
        tx.setTransform(mat);

        Group scene_root = new Group();
        scene_root.addChild(tx);

        // Sphere to render the shader onto
        GeometryData data = new GeometryData();
        data.geometryType = GeometryData.TRIANGLES;

        SphereGenerator generator = new SphereGenerator(1.0f, 64);
        generator.generate(data);

        TriangleArray geom = new TriangleArray();
        geom.setVertices(TriangleArray.COORDINATE_3,
                         data.coordinates,
                         data.vertexCount);

        String[] vert_shader_txt = { loadFile(VTX_SHADER_FILE) };
        String[] frag_shader_txt = { loadFile(FRAG_SHADER_FILE) };

        ShaderObject vert_shader = new ShaderObject(true);
        vert_shader.setSourceStrings(vert_shader_txt, 1);
        vert_shader.compile();

        ShaderObject frag_shader = new ShaderObject(false);
        frag_shader.setSourceStrings(frag_shader_txt, 1);
        frag_shader.compile();

        ShaderProgram shader_prog = new ShaderProgram();
        shader_prog.addShaderObject(vert_shader);
        shader_prog.addShaderObject(frag_shader);
        shader_prog.link();

        float[] light_dir = { 0.57735f, 0.57735f, 0.57735f, 0 };
        float[] h_vector = { 0.32506f, 0.32506f, 0.88808f, 0 };
        float[] ball_center = { 0, 0, 0, 1 };
        float[] specular_color = { 0.4f, 0.4f, 0.4f, 60 };
        float[] red =    { 0.6f, 0, 0, 1 };
        float[] blue =   { 0, 0.3f, 0.6f, 1 };
        float[] yellow = { 0.6f, 0.5f, 0, 1 };

        float[] hs_0 = { 1, 0, 0, 0, 0.2f };
        float[] hs_1 = { 0.309016994f, 0.951056516f, 0, 0.2f };
        float[] hs_2 = { -0.809016994f, 0.587785252f, 0, 0.2f };
        float[] hs_3 = { -0.809016994f, -0.587785252f, 0, 0.2f };
        float[] hs_4 = { 0.309016994f, -0.951056516f, 0, 0.2f };

        float[] in_out_init = { -3 };
        float[] stripe_width = { 0.3f };
        float[] f_width = { 0.005f };

        ShaderArguments shader_args = new ShaderArguments();
        shader_args.setUniform("LightDir", 4, light_dir, 1);
        shader_args.setUniform("HVector", 4, h_vector, 1);
        shader_args.setUniform("BallCenter", 4, ball_center, 1);
        shader_args.setUniform("SpecularColor", 4, specular_color, 1);
        shader_args.setUniform("Red", 4, red, 1);
        shader_args.setUniform("Blue", 4, blue, 1);
        shader_args.setUniform("Yellow", 4, yellow, 1);

        shader_args.setUniform("HalfSpace0", 4, hs_0, 1);
        shader_args.setUniform("HalfSpace1", 4, hs_1, 1);
        shader_args.setUniform("HalfSpace2", 4, hs_2, 1);
        shader_args.setUniform("HalfSpace3", 4, hs_3, 1);
        shader_args.setUniform("HalfSpace4", 4, hs_4, 1);

        shader_args.setUniform("InOrOutInit", 1, in_out_init, 1);
        shader_args.setUniform("StripeWidth", 1, stripe_width, 1);
        shader_args.setUniform("FWidth", 1, f_width, 1);

        GLSLangShader shader = new GLSLangShader();
        shader.setShaderProgram(shader_prog);
        shader.setShaderArguments(shader_args);

        Appearance app = new Appearance();
        app.setShader(shader);

        Shape3D shape = new Shape3D();
        shape.setGeometry(geom);
        shape.setAppearance(app);

        trans.set(0, 0, -0.2f);
        Matrix4f mat2 = new Matrix4f();
        mat2.setIdentity();
//        mat2.rotY((float)(Math.PI / 4));
//        mat2.setTranslation(trans);
//        mat2.setScale(0.05f);

        TransformGroup shape_tx = new TransformGroup();
        shape_tx.addChild(shape);
        shape_tx.setTransform(mat2);

        scene_root.addChild(shape_tx);

        Scene scene = new Scene();
        scene.setRenderedGeometry(scene_root);
        scene.setActiveView(vp);

        sceneManager.setScene(scene);
    }

    //---------------------------------------------------------------
    // Methods defined by Runnable
    //---------------------------------------------------------------

    /**
     * Run method for the shutdown hook. This is to deal with someone using
     * ctrl-C to kill the application. Makes sure that all the resources
     * are cleaned up properly.
     */
    public void run()
    {
        shutdownApp();
    }

    //---------------------------------------------------------------
    // Methods defined by WindowListener
    //---------------------------------------------------------------

    /**
     * Ignored
     */
    public void windowActivated(WindowEvent evt)
    {
    }

    /**
     * Ignored
     */
    public void windowClosed(WindowEvent evt)
    {
    }

    /**
     * Exit the application
     *
     * @param evt The event that caused this method to be called.
     */
    public void windowClosing(WindowEvent evt)
    {
        shutdownApp();
        System.exit(0);
    }

    /**
     * Ignored
     */
    public void windowDeactivated(WindowEvent evt)
    {
    }

    /**
     * Ignored
     */
    public void windowDeiconified(WindowEvent evt)
    {
    }

    /**
     * Ignored
     */
    public void windowIconified(WindowEvent evt)
    {
    }

    /**
     * When the window is opened, start everything up.
     */
    public void windowOpened(WindowEvent evt)
    {
        sceneManager.setEnabled(true);
    }

    //---------------------------------------------------------------
    // Local methods
    //---------------------------------------------------------------

    /**
     * Load the shader file. Find it relative to the classpath.
     *
     * @param file THe name of the file to load
     */
    private String loadFile(String name)
    {
        File file = new File(name);
        if(!file.exists())
        {
            System.out.println("Cannot find file " + name);
            return null;
        }

        String ret_val = null;

        try
        {
            FileReader is = new FileReader(file);
            StringBuffer buf = new StringBuffer();
            char[] read_buf = new char[1024];
            int num_read = 0;

            while((num_read = is.read(read_buf, 0, 1024)) != -1)
                buf.append(read_buf, 0, num_read);

            is.close();

            ret_val = buf.toString();
        }
        catch(IOException ioe)
        {
            System.out.println("I/O error " + ioe);
        }

        return ret_val;
    }

    //---------------------------------------------------------------
    // Local methods
    //---------------------------------------------------------------

    /**
     * Close down the application safely by destroying all the resources
     * currently in use.
     */
    private void shutdownApp()
    {
        sceneManager.setEnabled(false);
        surface.dispose();
    }

    public static void main(String[] args)
    {
        OBToyBallDemo demo = new OBToyBallDemo();
        demo.setVisible(true);
    }
}
